#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <float.h>
#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif
#include "linutil.h"


/**********************************************************************/
/*                         UTILITAIRES MATRICIELS                     */
/**********************************************************************/
/*---------------------------------------------------------------------
	Affection d'un scalaire aux composantes d'un vecteur de R3
	----------------------------------------------------------

	Entrees:
		double *u	: Adresse du v3 affect
		double s    : Scalaire  affecter

	Sorties:
		u est affect
		Retour		: Adresse du v3 affect

----------------------------------------------------------------------*/
double *setv3(double* u, double s)
{
    u[0] = s ;
    u[1] = s ;
    u[2] = s ;
	return u;
}

/*---------------------------------------------------------------------
	Recopie d'un vecteur de R3 : U = V
	----------------------------------

	Entrees:
		double *u	: Adresse du v3 affect
		double *v	: Adresse vecteur donn

	Sorties:
		u est affect
		Retour		: Adresse du v3 affect
----------------------------------------------------------------------*/
double *cpyv3(double* u, double* v)
{
    u[0] = v[0] ;
    u[1] = v[1] ;
    u[2] = v[2] ;
	return u;
}

/*---------------------------------------------------------------------
	Multiplication d'un vecteur de R3 par un scalaire : U = sV
	----------------------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du vecteur de R3 donn
		double s	: Scalaire donn

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V
----------------------------------------------------------------------*/
double *scav3(double* u, double s, double* v)
{
    u[0] = s * v[0] ;
    u[1] = s * v[1] ;
    u[2] = s * v[2] ;
	return u;
}

/*---------------------------------------------------------------------
	Addition de vecteurs de R3 : U = V + W
	--------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de R3 donn
		double *W	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *addv3(double* u, double* v, double* w)
{
    u[0] = v[0] + w[0] ;
    u[1] = v[1] + w[1] ;
    u[2] = v[2] + w[2] ;
	return u;
}

/*---------------------------------------------------------------------
	Soustraction de vecteurs de R3 : U = V - W
	------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de R3 donn
		double *W	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *subv3(double* u, double* v, double* w)
{
    u[0] = v[0] - w[0] ;
    u[1] = v[1] - w[1] ;
    u[2] = v[2] - w[2] ;
	return u;
}


/*---------------------------------------------------------------------
	Cumulation de vecteurs de R3 : U = V + aW
	-----------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de R3 donn
		double a	: Scalaire donn
		double *W	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *cumv3(double* u, double* v,double a, double* w)
{
    u[0] = v[0] + a*w[0] ;
    u[1] = v[1] + a*w[1] ;
    u[2] = v[2] + a*w[2] ;
	return u;
}

/*---------------------------------------------------------------------
	Produit vectoriel de vecteurs de R3 : A = B ^ C
	-----------------------------------------------

	Entrees:
		double *A	: Adresse du vecteur affect
		double *B	: Adresse du premier vecteur de R3 donn
		double *C	: Adresse du deuxime vecteur de R3 donn

	Sorties:
		A est affect
		Retour		: Adresse du vecteur affect

	Remarque : A peut occuper la mme mmoire que B ou C
----------------------------------------------------------------------*/
double *crossv3(double* a, double* b, double* c)
{
double p, q ;
      p  = b[1] * c[2] - b[2] * c[1];
      q  = b[2] * c[0] - b[0] * c[2];
    a[2] = b[0] * c[1] - b[1] * c[0];
    a[0] = p ;
    a[1] = q ;
	return a;
}


/*---------------------------------------------------------------------
	Longueur d'un vecteurs de R3 : s = sqrt(A'*A)
	---------------------------------------------

	Entrees:
		double *A	: Adresse vecteur donn

	Sorties:
		Retour		: Longueur du vecteur
----------------------------------------------------------------------*/
double longv3(double* a)
{
    return(sqrt(a[0]*a[0]+a[1]*a[1]+a[2]*a[2]));
}

/*---------------------------------------------------------------------
	Normalisation d'un vecteur de R3 : A = B / ||B||
	------------------------------------------------

	Entrees:
		double *A	: Adresse vecteur unitaire rsultat
		double *B	: Adresse vecteur donn

	Sorties:
		A est affect
		Retour		: Longueur du vecteur B

	Remarque : A et B peuvent occuper la mme mmoire
		       si ||B|| = 0  -> A[0] = 1, les autres nuls.

	Remarque : A peut occuper la mme mmoire que B
----------------------------------------------------------------------*/
double lunitv3(double* a, double* b)
{
    double l ;
    l = sqrt(b[0]*b[0]+b[1]*b[1]+b[2]*b[2]);
    if (l == 0.)
    {
        a[0]=1.; a[1]=0.; a[2]=0.;
    }
    else
    {
        a[0] = b[0]/l;
		a[1] = b[1]/l;
        a[2] = b[2]/l;
    }
    return(l);
}


/*---------------------------------------------------------------------
	Matrice Identite de R3 :  A = I
	-------------------------------

	Entrees:
		double *A	: Adresse de la matrice affecte

    Sorties:
		A est affecte
		Retour		: Adresse de A
----------------------------------------------------------------------*/
double *eyem3(double *a)
{
    a[0] = a[4] = a[8] = 1. ;
    a[1] = a[2] = a[3] = 0. ;
    a[5] = a[6] = a[7] = 0. ;
	return a ;
}

/*---------------------------------------------------------------------
	Produit matrice x matrice de R3 : A = B C
	-----------------------------------------------

	Entrees:
		double *A	: Adresse de la matrice affecte
		double *B	: Adresse de la premire matrice de R3 donne
		double *C	: Adresse de la deuxime matrice de R3 donne

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice affecte

    Remarque : A peut occuper la mme mmoire que B ou C
----------------------------------------------------------------------*/
double *m3m3(double* a,double* b, double* c)
{
register int i ;
double d[8] ;
    d[0] = b[0] * c[0] + b[3] * c[1] + b[6] * c[2];
    d[1] = b[1] * c[0] + b[4] * c[1] + b[7] * c[2];
    d[2] = b[2] * c[0] + b[5] * c[1] + b[8] * c[2];

    d[3] = b[0] * c[3] + b[3] * c[4] + b[6] * c[5];
    d[4] = b[1] * c[3] + b[4] * c[4] + b[7] * c[5];
    d[5] = b[2] * c[3] + b[5] * c[4] + b[8] * c[5];

    d[6] = b[0] * c[6] + b[3] * c[7] + b[6] * c[8];
    d[7] = b[1] * c[6] + b[4] * c[7] + b[7] * c[8];
    a[8] = b[2] * c[6] + b[5] * c[7] + b[8] * c[8];
    for(i = 0 ; i < 8 ; ++i) a[i] = d[i] ;
	return a;
}

/*---------------------------------------------------------------------
	Produit matrice x matrice transpose de R3 : A = B C'
	----------------------------------------------------

	Entrees:
		double *A	: Adresse de la matrice affecte
		double *B	: Adresse de la premire matrice de R3 donne
		double *C	: Adresse de la deuxime matrice de R3 donne

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice affecte

    Remarque : A peut occuper la mme mmoire que B ou C
----------------------------------------------------------------------*/
double *m3tm3(double* a, double* b, double* c)
{
register int i ;
double d[8] ;
    d[0] = b[0] * c[0] + b[3] * c[3] + b[6] * c[6];
    d[1] = b[1] * c[0] + b[4] * c[3] + b[7] * c[6];
    d[2] = b[2] * c[0] + b[5] * c[3] + b[8] * c[6];

    d[3] = b[0] * c[1] + b[3] * c[4] + b[6] * c[7];
    d[4] = b[1] * c[1] + b[4] * c[4] + b[7] * c[7];
    d[5] = b[2] * c[1] + b[5] * c[4] + b[8] * c[7];

    d[6] = b[0] * c[2] + b[3] * c[5] + b[6] * c[8];
    d[7] = b[1] * c[2] + b[4] * c[5] + b[7] * c[8];
    a[8] = b[2] * c[2] + b[5] * c[5] + b[8] * c[8];
    for(i = 0 ; i < 8 ; ++i) a[i] = d[i] ;
	return a;
}

/*---------------------------------------------------------------------
	Produit matrice x vecteurs de R3 : w = A v
	-----------------------------------------------

	Entrees:
		double *w	: Adresse du vecteur affect
		double *A	: Adresse de la matrice de R3 donn
		double *v	: Adresse du vecteur de R3 donn

	Sorties:
		w est affect
		Retour		: Adresse du vecteur affect

    Remarque : w et v peuvent occuper la mme mmoire.
----------------------------------------------------------------------*/
double *m3v3(double* w, double* a, double* v)
{
double p, q ;
      p  = a[0] * v[0] + a[3] * v[1] + a[6] * v[2];
      q  = a[1] * v[0] + a[4] * v[1] + a[7] * v[2];
    w[2] = a[2] * v[0] + a[5] * v[1] + a[8] * v[2];
    w[0] = p ;
    w[1] = q ;
	return w;
}

/*---------------------------------------------------------------------
	Produit matrice transpose x vecteurs de R3 : w = A' v
	------------------------------------------------------

	Entrees:
		double *w	: Adresse du vecteur affect
		double *A	: Adresse de la matrice de R3 donn
		double *v	: Adresse du vecteur de R3 donn

	Sorties:
		w est affect
		Retour		: Adresse du vecteur affect

    Remarque : w et v peuvent occuper la mme mmoire.
----------------------------------------------------------------------*/
double *tm3v3(double* w, double* a, double* v)
{
double p, q ;
      p  = a[0] * v[0] + a[1] * v[1] + a[2] * v[2];
      q  = a[3] * v[0] + a[4] * v[1] + a[5] * v[2];
    w[2] = a[6] * v[0] + a[7] * v[1] + a[8] * v[2];
    w[0] = p ;
    w[1] = q ;
	return w;
}

/*---------------------------------------------------------------------
	Produit matrice transpose x matrice de R3 : A = B'C
	----------------------------------------------------

	Entrees:
		double *A	: Adresse de la matrice affecte
		double *B	: Adresse de la premire matrice de R3 donne
		double *C	: Adresse de la deuxime matrice de R3 donne

	Sorties:
		A est affecte
		Retour		: Adresse de la matrice affecte

    Remarque : A peut occuper la mme mmoire que B ou C
----------------------------------------------------------------------*/
double *tm3m3( double* a,double* b, double* c)
{
register int i;
double d[8] ;
    d[0] = b[0] * c[0] + b[1] * c[1] + b[2] * c[2];
    d[1] = b[3] * c[0] + b[4] * c[1] + b[5] * c[2];
    d[2] = b[6] * c[0] + b[7] * c[1] + b[8] * c[2];

    d[3] = b[0] * c[3] + b[1] * c[4] + b[2] * c[5];
    d[4] = b[3] * c[3] + b[4] * c[4] + b[5] * c[5];
    d[5] = b[6] * c[3] + b[7] * c[4] + b[8] * c[5];

    d[6] = b[0] * c[6] + b[1] * c[7] + b[2] * c[8];
    d[7] = b[3] * c[6] + b[4] * c[7] + b[5] * c[8];
    a[8] = b[6] * c[6] + b[7] * c[7] + b[8] * c[8];
    for(i = 0 ; i < 8 ; ++i) a[i] = d[i] ;
	return a;
}


/*---------------------------------------------------------------------
	Gnration matrice de rotation d'angle a autour de Ox
	-----------------------------------------------------

	Entrees:
		double *m	: Adresse de la matrice affecte
		double a	: Angle de rotation (radians)

	Sorties:
		m est affecte
		Retour		: Adresse de la matrice affecte
----------------------------------------------------------------------*/
double *mrotx(double *m, double a)
{
double s = sin(a), c = cos(a);
	m[0] = 1.; m[3] = 0.; m[6] = 0.;
	m[1] = 0.; m[4] = c ; m[7] =-s ;
	m[2] = 0.; m[5] = s ; m[8] = c ;
    return m ;
}

/*---------------------------------------------------------------------
	Gnration matrice de rotation d'angle a autour de Oz
	-----------------------------------------------------

	Entrees:
		double *m	: Adresse de la matrice affecte
		double a	: Angle de rotation  (radians)

	Sorties:
		m est affecte
		Retour		: Adresse de la matrice affecte
----------------------------------------------------------------------*/
double *mrotz(double *m, double a)
{
double s = sin(a), c = cos(a);
	m[0] = c ; m[3] =-s ; m[6] = 0.;
	m[1] = s ; m[4] = c ; m[7] = 0.;
	m[2] = 0.; m[5] = 0.; m[8] = 1.;
    return m ;
}

/*---------------------------------------------------------------------
	Gnration matrice de rotation d'angle a autour de Oy
	-----------------------------------------------------

	Entrees:
		double *m	: Adresse de la matrice affecte
		double a	: Angle de rotation  (radians)

	Sorties:
		m est affecte
		Retour		: Adresse de la matrice affecte
----------------------------------------------------------------------*/
double *mroty(double *m, double a)
{
double s = sin(a), c = cos(a);
	m[0] = c ; m[3] = 0.; m[6] = s ;
	m[1] = 0.; m[4] = 1.; m[7] = 0.;
	m[2] =-s ; m[5] = 0.; m[8] = c ;
    return m ;
}
/*---------------------------------------------------------------------
	Gnration matrice de rotation d'angle a autour du vecteur V
	------------------------------------------------------------

	Entrees:
		double *m	: Adresse de la matrice affecte
		double a	: Angle de rotation  (radians)
		double *v	: Adresse du vecteur unitaire port
		              par l'axe de rotation

	Sorties:
		m est affecte
		Retour		: Adresse de la matrice affecte
----------------------------------------------------------------------*/
double *av3tomrot(double* m, double a, double* v)
{
    double wx, wy, wz;
    double k;

    wx = v[0];
    wy = v[1];
    wz = v[2];
    k = 1.-cos(a);
    m[0] = 1. + k * (wx * wx - 1.);
    m[4] = 1. + k * (wy * wy - 1.);
    m[8] = 1. + k * (wz * wz - 1.);
    m[1] = k * wx * wy;
    m[2] = k * wz * wx;
    m[5] = k * wy * wz;
    k = sin(a);
    wx = k * wx;
    wy = k * wy;
    wz = k * wz;
    m[7] = m[5] - wx;
    m[5] += wx;
    m[6] = m[2] + wy;
    m[2] -= wy;
    m[3] = m[1] - wz;
    m[1] += wz;
	return m;
}

/*----------------------------------------------------------------------
	Extraction des angles l,t,r (/Z0,/X1,/Y2) d'une matrice de rotation
	--------------------------------------------------------------------

	Entrees:
		double *ltr	: Adresse des angles l,t,r  (radians)
		double *A	: Adresse de la matrice de rotation donn

	Sorties:
		angles ltr calculs (radians)
		Retour		: Adresse des angles l,t,r

    Remarque :
		Pour abs(t) > pi/2 - 1.E-6   conservation de la somme ou de la
		difference des valeurs ltr passes en rfrence.
----------------------------------------------------------------------*/
double *mrot2ltr (double *ltr, double *A)
{
double l, t, r, som,dif,pis2 ;

	pis2 = M_PI/2. ;
	t = asin( A[5] ) ;
	if ( fabs(t) < pis2 - 1.E-6)
	{
	   l = atan2( -A[3] , A[4] ) ;
	   r = atan2( -A[2] , A[8] ) ;
	}
	else
	{
	   if ( t > 0.)
	   {
	      som = atan2( A[1],A[0] ) ;
	      dif = ltr[0] - ltr[2] ;
	   }
	   else
	   {
	      dif = atan2( A[1],A[0] ) ;
	      som = ltr[0] + ltr[2] ;
	   }
	   l = (som + dif)/2. ;
	   r = (som - dif)/2. ;
	}

	if(fabs(ltr[1]) > pis2)
	{
	   if(l < 0.) l = M_PI + l ; else l = -M_PI + l ;
	   if(t > 0.) t = M_PI - t ; else t = -M_PI - t ;
	   if(r < 0.) r = M_PI + r ; else r = -M_PI + r ;
	}
	ltr[0] = l ;
	ltr[1] = t ;
	ltr[2] = r ;

	return ltr;
}

/*---------------------------------------------------------------------
	Produit Por x Por de R3 : A = B C
	---------------------------------

	Entrees:
		double *A	: Adresse du Por affect
		double *B	: Adresse du premier Por donn
		double *C	: Adresse du deuxime Por donn

	Sorties:
		A est affect
		Retour		: Adresse du Por affect

    Remarque : A peut occuper la mme mmoire que B ou C
----------------------------------------------------------------------*/
double *porpor(double* a, double* b, double* c)
{
double d[3] ;
    m3v3(d, b+3, c);
    m3m3(a+3, b+3, c+3);
    addv3(a, b, d) ;
	return a;
}

/*---------------------------------------------------------------------
	Produit Por x Por inverse de R3 : A = B inv(C)
	----------------------------------------------

	Entrees:
		double *A	: Adresse du Por affect
		double *B	: Adresse du premier Por donn
		double *C	: Adresse du deuxime Por donn

	Sorties:
		A est affect
		Retour		: Adresse du Por affect

    Remarque : A peut occuper la mme mmoire que B ou C
----------------------------------------------------------------------*/
double *porpori(double* a, double* b, double* c)
{
double d[3] ;
    m3tm3(a+3, b+3 ,c+3) ;
    m3v3(d, a+3, c) ;
    subv3(a, b, d) ;
	return a;
}


/*---------------------------------------------------------------------
	Affectation d'un scalaire aux composantes vecteur de Rn : Ui = s
	----------------------------------------------------------------

	Entrees:
		double *u	: Adresse du vecteur affect
		double s	: scalaire donn
		int n		: dimension de v

	Sorties:
		u est affect
		Retour		: Adresse du vecteur affect
----------------------------------------------------------------------*/
double *setvec (double* u, double s, int n)
{
    if (n > 0) while (n--) u[n] = s;
	return u;
}

/*---------------------------------------------------------------------
	Recopie d'un vecteur de Rn : U = V
	----------------------------------

	Entrees:
		double *u	: Adresse du vecteur affect
		double *v	: Adresse vecteur donn
		int n		: dimension de v

	Sorties:
		u est affect
		Retour		: Adresse du vecteur affect
----------------------------------------------------------------------*/
double *cpyvec (double* u, double* v, int n)
{
    if (n > 0) while (n--) u[n] = v[n];
	return u;
}


/*---------------------------------------------------------------------
	Addition de vecteurs de Rn : U = V + W
	--------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de Rn donn
		double *W	: Adresse du deuxime vecteur de Rn donn
		int n		: dimension des vecteurs

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *addvec (double* u, double* v, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] + w[n];
	return u;
}

/*---------------------------------------------------------------------
	Soustraction de vecteurs de Rn : U = V - W
	------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de Rn donn
		double *W	: Adresse du deuxime vecteur de Rn donn
		int n		: dimension des vecteurs

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W
----------------------------------------------------------------------*/
double *subvec (double* u, double* v, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] - w[n];
	return u;
}


/*---------------------------------------------------------------------
	Produit scalaire de vecteurs de Rn : s = <U . V>
	------------------------------------------------

	Entrees:
		double *U	: Adresse du premier vecteur de Rn donn
		double *V	: Adresse du deuxime vecteur de Rn donn
		int n		: dimension des vecteurs

	Sorties:
		Retour		: produit scalaire
----------------------------------------------------------------------*/
double dotvec (double* u, double* v, int n)
{
double s ;
    s = 0. ;
    if (n > 0) while (n--) s += u[n] * v[n] ;
    return(s);
}

/*---------------------------------------------------------------------
	Division terme  terme de vecteurs de Rn : U = V / W
	----------------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double *V	: Adresse du premier vecteur de Rn donn
		double *W	: Adresse du deuxime vecteur de Rn donn
		int n		: dimension des vecteurs

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V ou W

    ATTENTION : DIVISION PAR ZERO NON TESTEE !!!!!!!!!!!!!!!
----------------------------------------------------------------------*/
double *divvec (double* u, double* v, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] / w[n];
	return u;
}

/*---------------------------------------------------------------------
    Valeur absolue d'un vecteurs de Rn : U = |V|
	--------------------------------------------

	Entrees:
		double *u	: Adresse du vecteur affect
		double *v	: Adresse vecteur donn
		int n		: dimension des vecteurs

	Sorties:
		u est affect
		Retour		: Adresse du vecteur affect

    Remarque : u et v peuvent occuper la mme mmoire.
----------------------------------------------------------------------*/
double *absvec (double* u, double* v, int n)
{
    if (n > 0) while (n--) u[n] = fabs(v[n]);
	return u;
}

/*---------------------------------------------------------------------
	Composante maximale d'un vecteur de Rn : s = max(Ui)
	-----------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn
		int n		: dimension du vecteur


	Sorties:
		Retour		: Composante maximale du vecteur
----------------------------------------------------------------------*/
double xmaxvec (double* u, int n)
{
double s = -DBL_MAX;
    if (n > 0) while (n--) if (u[n]>s) s = u[n];
	return s;
}

/*---------------------------------------------------------------------
	index du + grand lment en valeur absolue (0  n-1)
	----------------------------------------------------

	Entrees:
		double *u	: Adresse vecteur donn
		int n		: dimension du vecteur

	Sorties:
		Retour		: index
----------------------------------------------------------------------*/
int imaxavec (double *u, int n)
{
int i, id ;
double dmax, a ;

    if (n <= 1) return (n-1) ;
    dmax = fabs(u[0]) ; id = 0 ;
    for (i = 1 ; i < n ; i++)
    { a = fabs(u[i]) ; if (a > dmax) {id = i ; dmax = a ; } }
    return id ;
}

/*---------------------------------------------------------------------
	Multiplication d'un vecteur de Rn par un scalaire : U = aV
	----------------------------------------------------------

	Entrees:
		double *U	: Adresse du vecteur affect
		double a	: Scalaire donn
		double *V	: Adresse du vecteur de Rn donn
		int n		: dimension du vecteur

	Sorties:
		U est affect
		Retour		: Adresse du vecteur affect

	Remarque : U peut occuper la mme mmoire que V
----------------------------------------------------------------------*/
double *scavec (double *u, double a, double *v, int n)
{
    if (n > 0) while (n--) u[n] = a * v[n] ;
	return u;
}

/*---------------------------------------------------------------------
	Produit de matrice : Amp = Bmn x [C']np
	---------------------------------------

	Entrees:
		double *A	: Adresse de A
		double *B	: Adresse de B
		double *C	: Adresse de C
		int    m         : Nombre de lignes de A et de lignes de B
		int    n         : Nombre de colonnes de B et de colonnes de C
		int    p         : Nombre de colonnes de A et de lignes de C

    Sortie :
		A est calcule
		Retour	: Adresse de A			:

    Remarque : premire et dernire dimension : celles du rsultat
	A ne doit pas occuper la mme zone mmoire que B ou C
----------------------------------------------------------------------*/
double *mattmat (double *a, double *b, double *c, int m, int n, int p)
{
int i, j, k, ij, ik, jk ;

    ij = -1 ;
    for (j = 0 ; j < p ; j++)
	for (i = 0 ; i < m ; i++)
	{
	    a[++ij] = 0. ;
		for (k = 0, ik = i, jk = j ; k < n ; k++, ik += m, jk += p)
		    a[ij] += b[ik]*c[jk] ;
	}
	return a;
}

/*---------------------------------------------------------------------
	Produit de matrice : Amp = [B']mn x Cnp
	---------------------------------------

	Entrees:
		double *A	: Adresse de A
		double *B	: Adresse de B
		double *C	: Adresse de C
		int    m         : Nombre de lignes de A et de colonnes de B
		int    n         : Nombre de lignes de B et  de C
		int    p         : Nombre de colonnes de A et de C

    Sortie :
		A est calcule
		Retour	: Adresse de A			:

    Remarque : premire et dernire dimension : celles du rsultat
	A ne doit pas occuper la mme zone mmoire que B ou C
----------------------------------------------------------------------*/
double *tmatmat (double *a, double *b, double *c, int m, int n, int p)
{
int i, j, k, ij, ki, kj, kj0 ;

    for (j = 0, ij = 0, kj0 = 0 ; j < p ; j++, kj0 += n)
    {
	for (i = 0, ki = 0 ; i < m ; i++, ij++)
	{
	    a[ij] = 0. ;
	    for (k = 0, kj = kj0 ; k < n ; k++) a[ij] += b[ki++] * c[kj++] ;
	}
    }
	return a;
}

/*--------------------------------------------------------------------*/
static void cum1_vect (double *y, double a, double *x, int n)
{
/* Cumulation de vecteurs */
    if (n > 0 && a != 0.) while (n--) y[n] += a * x[n] ;
}
/*-------------------------------------------------------------------------*/
static void fact1_vect (double *x, double a, int n)
{ /* multiplie un vecteur par une constante */

    if (n > 0) while (n--) x[n] = a * x[n] ;
}
/*-------------------------------------------------------------------------*/
static int lufactor (double *a, int lda, int n, int *ipvt)
{
/*
     lufactor effectue la factorisation L U d'une matrice n x n
     extraite d'une matrice rangees par vecteurs de lda composantes.
     Les facteurs sont renvoyes dans A et les indices des pivots
     des lignes dans ipvt.
     Renvoi 0 si OK ou l'indice fortran du 1er element nul de la diag

     D'apres : LINPACK. 08/14/78 CLEVE MOLER, UNIVERSITY OF NEW MEXICO
     ARGONNE NATIONAL LAB.
     Adapte en C par MLL le 28/04/92

     Utilise cum1_vect, fact1_vect
*/

double t ;
int info, j, k, l, nm1, lk, kk, lj, kj, klda;
int lda1 = 1 + lda ;

/*  Elimination de Gauss avec pivotage partiel */

    info = 0 ;
    nm1 = n - 1 ;
    kk = 0 ;
    if (nm1 >= 1)
    {
	for (k = 0 , klda = 0 , kk = 0 ; k < nm1 ;
					 k++ , klda += lda, kk += lda1)
	{
	/*  Recherche l = indice du pivot */
	    l = imaxavec (&a[kk], n-k) + k ;
	    ipvt[k] = l ;
	/* Si 0 la colonne est deja triangularisee */
	    lk = l + klda ;
	    if ( a[lk] != 0.)
	    {
		if (l != k) { t = a[lk] ; a[lk] = a[kk] ; a[kk] = t ; }
	    /* Calcul des multiplicateurs */
		t = -1./a[kk] ;
		fact1_vect (&a[kk+1], t, nm1-k) ;
	    /* Elimination des colonnes */
		lj = lk ; kj = kk ;
		for (j = k+1 ; j < n ; j++)
		{
		    lj += lda ; kj += lda ; t= a[lj] ;
		    if (l != k) { a[lj] = a[kj] ; a[kj] = t ; }
		    cum1_vect (&a[kj+1], t, &a[kk+1], nm1-k) ;
		}
	    }
	    else info = k+1 ;
	}
    }
    ipvt[nm1] = nm1 ;
    kk = nm1*lda1 ;
    if ( a[kk] == 0.) info = n ;
    return info ;
}
/*-------------------------------------------------------------------------*/
static int lusolv (double *a, int lda, int n, int *ipvt, double *b, int trp)
{
/*
      Renvoi 1 si Ok et 0 si tentative division par 0 (calcul non fait)
      LINPACK. THIS VERSION DATED 08/14/78 .
      CLEVE MOLER, UNIVERSITY OF NEW MEXICO, ARGONNE NATIONAL LAB.
      Adapte en C par MLL le 28/04/92
*/
double t, diag ;
int k, kp1, klda, l, nm1 ;

    nm1 = n - 1 ;
    if (trp == 0 )
    {
/*
         trp = 0 , resout  A * X = B
         resout d'abord  L*Y = B
*/
	if (nm1 > 0)
        {
            for (k = 0 , klda = 0 ; k < nm1 ; k++, klda += lda )
            {
                kp1 = k + 1 ;
                l = ipvt[k] ; t = b[l] ;
		if (l != k) { b[l] = b[k] ; b[k] = t ; }
		cum1_vect (&b[kp1], t, &a[klda+kp1], n-k-1) ;
            }
        }
/*
         resout  U*X = Y
*/
        for (k = nm1 , klda = nm1 * lda ; k >= 0 ; k--, klda -= lda)
	{
	    if ( (diag = a[k+klda]) == 0.) return 0 ;
            b[k] /= diag ;
	    t = - b[k] ;
	    cum1_vect (b, t , &a[klda], k) ;
        }
    }
    else
    {
/*
         trp = 1, resout  A' * X = B
         resout d'abord   U' * Y = B
*/
        for (k = 0, klda = 0 ; k < n ; k++, klda += lda)
        {
	    t = dotvec (&a[klda], b, k) ;
            if ( (diag = a[k+klda]) == 0.) return 0 ;
            b[k] = ( b[k] - t ) / diag ;
        }
/*
         resout L' * X = Y
*/
	if (nm1 > 0)
	{
	    for (k = n-2, klda = k * lda ; k >= 0 ; k-- , klda -= lda)
	    {
		kp1 = k + 1 ;
		b[k] += dotvec (&a[klda+kp1], &b[kp1], nm1 - k) ;
		l = ipvt[k] ;
		if (l != k) {t = b[l]; b[l] = b[k] ; b[k] = t ; }
	    }
	}
    }
    return 1 ;
}
/*---------------------------------------------------------------------
	               /  Y = A  * X  (si trp = 0)  -> X = inv(A)  * Y
	Rsolution de  |
	               \  Y = A' * X  (si trp = 1)  -> X = inv(A') * Y
	-------------------------------------------------------------------

    !!!!!!!!!!  ATTENTION : LA MATRICE A EST DETRUITE  !!!!!!!!!!!

	Entrees:
		double *X		: Adresse de X vecteurs affects
		double *A		: Adresse de A (qui sera dtruite)
		double *Y		: Adresse de Y vecteurs donns
		int    n		: dimension de A (n x n)
		int    ns		: Nombre de colonnes de Y (et de X)
		int    trp		: Flag de transposition de A

    Sortie :
		X est calcule
		Retour			: 1 si calcul de X effectu
		                  0 si rsolution impossible
----------------------------------------------------------------------*/
int solvmat (double *X, double *A, double *Y, int n, int ns, int trp)
{
#define DIMENSION_MAX 100
int flag, ipvt[DIMENSION_MAX] ;

    if (n > DIMENSION_MAX)
    {
		puts ("DIMENSION_MAX insuffisante dans solvmat");
	    puts("Modifier et recompiler") ;
		exit(1) ;
	}

	if (X != Y) cpyvec(X, Y, n*ns) ;

    if ( lufactor (A, n, n, ipvt) == 0)
    {
	flag = 1 ;
	while (ns--) lusolv (A, n, n, ipvt, &X[n*ns], trp) ;
    }
    else flag = 0 ;

    return flag ;
}
